'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var jsCore = require('@lorisleiva/js-core');
var beet = require('@metaplex-foundation/beet');
var beetSolana = require('@metaplex-foundation/beet-solana');
var web3_js = require('@solana/web3.js');
var buffer = require('buffer');
var errors = require('./errors.cjs');
var numbers = require('./numbers.cjs');

function _interopNamespace(e) {
  if (e && e.__esModule) return e;
  var n = Object.create(null);
  if (e) {
    Object.keys(e).forEach(function (k) {
      if (k !== 'default') {
        var d = Object.getOwnPropertyDescriptor(e, k);
        Object.defineProperty(n, k, d.get ? d : {
          enumerable: true,
          get: function () { return e[k]; }
        });
      }
    });
  }
  n["default"] = e;
  return Object.freeze(n);
}

var beet__namespace = /*#__PURE__*/_interopNamespace(beet);
var beetSolana__namespace = /*#__PURE__*/_interopNamespace(beetSolana);

class BeetSerializer {
  tuple(items, description) {
    const itemDescriptions = items.map(item => item.description).join(', ');
    return {
      description: description ?? `tuple(${itemDescriptions})`,
      serialize: value => {
        if (value.length !== items.length) {
          throw new Error(`Expected tuple to have ${items.length} items but got ${value.length}.`);
        }
        return jsCore.mergeBytes(items.map((item, index) => item.serialize(value[index])));
      },
      deserialize: (bytes, offset = 0) => {
        const values = [];
        items.forEach(serializer => {
          const [newValue, newOffset] = serializer.deserialize(bytes, offset);
          values.push(newValue);
          offset = newOffset;
        });
        return [values, offset];
      }
    };
  }
  vec(itemSerializer, description) {
    return {
      description: description ?? `vec(${itemSerializer.description})`,
      serialize: value => {
        const lengthBytes = numbers.u32().serialize(value.length);
        const itemBytes = value.map(item => itemSerializer.serialize(item));
        return jsCore.mergeBytes([lengthBytes, ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const values = [];
        const [length, newOffset] = numbers.u32().deserialize(bytes, offset);
        offset = newOffset;
        for (let i = 0; i < length; i += 1) {
          const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
          values.push(value);
          offset = newOffset;
        }
        return [values, offset];
      }
    };
  }
  array(itemSerializer, size, description) {
    return {
      description: description ?? `array(${itemSerializer.description}; ${size})`,
      serialize: value => {
        if (value.length !== size) {
          throw new Error(`Expected array to have ${size} items but got ${value.length}.`);
        }
        return jsCore.mergeBytes(value.map(item => itemSerializer.serialize(item)));
      },
      deserialize: (bytes, offset = 0) => {
        const values = [];
        for (let i = 0; i < size; i += 1) {
          const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
          values.push(value);
          offset = newOffset;
        }
        return [values, offset];
      }
    };
  }
  map(keySerializer, valueSerializer, description) {
    return {
      description: description ?? `map(${keySerializer.description}, ${valueSerializer.description})`,
      serialize: map => {
        const lengthBytes = numbers.u32().serialize(map.size);
        const itemBytes = Array.from(map, ([key, value]) => jsCore.mergeBytes([keySerializer.serialize(key), valueSerializer.serialize(value)]));
        return jsCore.mergeBytes([lengthBytes, ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const map = new Map();
        const [length, newOffset] = numbers.u32().deserialize(bytes, offset);
        offset = newOffset;
        for (let i = 0; i < length; i += 1) {
          const [key, kOffset] = keySerializer.deserialize(bytes, offset);
          offset = kOffset;
          const [value, vOffset] = valueSerializer.deserialize(bytes, offset);
          offset = vOffset;
          map.set(key, value);
        }
        return [map, offset];
      }
    };
  }
  set(itemSerializer, description) {
    return {
      description: description ?? `set(${itemSerializer.description})`,
      serialize: set => {
        const lengthBytes = numbers.u32().serialize(set.size);
        const itemBytes = Array.from(set, value => itemSerializer.serialize(value));
        return jsCore.mergeBytes([lengthBytes, ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const set = new Set();
        const [length, newOffset] = numbers.u32().deserialize(bytes, offset);
        offset = newOffset;
        for (let i = 0; i < length; i += 1) {
          const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
          offset = newOffset;
          set.add(value);
        }
        return [set, offset];
      }
    };
  }
  option(itemSerializer, description) {
    return {
      description: description ?? `option(${itemSerializer.description})`,
      serialize: option => {
        const prefixByte = numbers.bool().serialize(option !== null);
        const itemBytes = option !== null ? itemSerializer.serialize(option) : new Uint8Array();
        return jsCore.mergeBytes([prefixByte, itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const [isSome, prefixOffset] = numbers.bool().deserialize(bytes, offset);
        offset = prefixOffset;
        if (!isSome) {
          return [null, offset];
        }
        const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
        offset = newOffset;
        return [value, offset];
      }
    };
  }
  struct(fields, description) {
    const fieldDescriptions = fields.map(([name, serializer]) => `${String(name)}: ${serializer.description}`).join(', ');
    return {
      description: description ?? `struct(${fieldDescriptions})`,
      serialize: struct => {
        const fieldBytes = fields.map(([key, serializer]) => serializer.serialize(struct[key]));
        return jsCore.mergeBytes(fieldBytes);
      },
      deserialize: (bytes, offset = 0) => {
        const struct = {};
        fields.forEach(([key, serializer]) => {
          const [value, newOffset] = serializer.deserialize(bytes, offset);
          offset = newOffset;
          struct[key] = value;
        });
        return [struct, offset];
      }
    };
  }
  enum(constructor, description) {
    const enumValues = Object.values(constructor);
    const isNumericEnum = enumValues.some(v => typeof v === 'number');
    const valueDescriptions = enumValues.filter(v => typeof v === 'string').join(', ');
    function getVariantKeyValue(value) {
      if (typeof value === 'number') {
        return [enumValues[value], value];
      }
      const variantValue = constructor[value];
      if (typeof variantValue === 'number') {
        return [value, variantValue];
      }
      const indexOfValue = enumValues.indexOf(variantValue);
      if (indexOfValue >= 0) {
        return [variantValue, indexOfValue];
      }
      return [value, enumValues.indexOf(value)];
    }
    function checkVariantExists(variantKey) {
      if (!enumValues.includes(variantKey)) {
        throw new Error(`Invalid enum variant. Got "${variantKey}", expected one of ` + `[${enumValues.join(', ')}]`);
      }
    }
    return {
      description: description ?? `enum(${valueDescriptions})`,
      serialize: value => {
        const [variantKey, variantValue] = getVariantKeyValue(value);
        checkVariantExists(variantKey);
        return numbers.u8().serialize(variantValue);
      },
      deserialize: (bytes, offset = 0) => {
        const [value, newOffset] = numbers.u8().deserialize(bytes, offset);
        offset = newOffset;
        const [variantKey, variantValue] = getVariantKeyValue(value);
        checkVariantExists(variantKey);
        return [isNumericEnum ? variantValue : variantKey, offset];
      }
    };
  }
  dataEnum(fields, description) {
    const fieldDescriptions = fields.map(([name, serializer]) => `${String(name)}${serializer ? `: ${serializer.description}` : ''}`).join(', ');
    return {
      description: description ?? `dataEnum(${fieldDescriptions})`,
      serialize: variant => {
        const discriminator = fields.findIndex(([key]) => variant.__kind === key);
        if (discriminator < 0) {
          throw new Error(`Invalid data enum variant. Got "${variant.__kind}", expected one of ` + `[${fields.map(([key]) => key).join(', ')}]`);
        }
        const variantPrefix = numbers.u8().serialize(discriminator);
        const variantSerializer = fields[discriminator][1];
        const variantBytes = variantSerializer.serialize(variant);
        return jsCore.mergeBytes([variantPrefix, variantBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const [discriminator, dOffset] = numbers.u8().deserialize(bytes, offset);
        offset = dOffset;
        const variantField = fields[discriminator] ?? null;
        if (!variantField) {
          throw new Error(`Data enum index "${discriminator}" is out of range. ` + `Index should be between 0 and ${fields.length - 1}.`);
        }
        const [variant, vOffset] = variantField[1].deserialize(bytes, offset);
        offset = vOffset;
        return [{
          __kind: variantField[0],
          ...(variant ?? {})
        }, offset];
      }
    };
  }
  get unit() {
    return {
      description: 'unit',
      serialize: () => new Uint8Array(),
      deserialize: (_bytes, offset = 0) => [undefined, offset]
    };
  }
  get bool() {
    return numbers.bool();
  }
  get u8() {
    return numbers.u8();
  }
  get u16() {
    return numbers.u16();
  }
  get u32() {
    return numbers.u32();
  }
  get u64() {
    return numbers.u64();
  }
  get u128() {
    return numbers.u128();
  }
  get i8() {
    return numbers.i8();
  }
  get i16() {
    return numbers.i16();
  }
  get i32() {
    return numbers.i32();
  }
  get i64() {
    return numbers.i64();
  }
  get i128() {
    return numbers.i128();
  }
  get f32() {
    return {
      description: 'f32 [not supported]',
      serialize: () => {
        throw new errors.OperationNotSupportedError('f32');
      },
      deserialize: () => {
        throw new errors.OperationNotSupportedError('f32');
      }
    };
  }
  get f64() {
    return {
      description: 'f64 [not supported]',
      serialize: () => {
        throw new errors.OperationNotSupportedError('f64');
      },
      deserialize: () => {
        throw new errors.OperationNotSupportedError('f64');
      }
    };
  }
  get string() {
    return {
      description: 'string',
      serialize: value => {
        const stringBeet = beet__namespace.utf8String.toFixedFromValue(value);
        const buffer$1 = buffer.Buffer.alloc(stringBeet.byteSize);
        stringBeet.write(buffer$1, 0, value);
        return new Uint8Array(buffer$1);
      },
      deserialize: (bytes, offset = 0) => {
        const buffer$1 = buffer.Buffer.from(bytes);
        const stringBeet = beet__namespace.utf8String.toFixedFromData(buffer$1, offset);
        const value = stringBeet.read(buffer$1, offset);
        return [value, offset + stringBeet.byteSize];
      }
    };
  }
  get bytes() {
    return {
      description: 'bytes',
      serialize: value => new Uint8Array(value),
      deserialize: (bytes, offset = 0) => [new Uint8Array(bytes), offset + bytes.length]
    };
  }
  get publicKey() {
    return {
      description: 'publicKey',
      serialize: value => {
        const buffer$1 = buffer.Buffer.alloc(beetSolana__namespace.publicKey.byteSize);
        beetSolana__namespace.publicKey.write(buffer$1, 0, new web3_js.PublicKey(value));
        return new Uint8Array(buffer$1);
      },
      deserialize: (bytes, offset = 0) => {
        const buffer$1 = buffer.Buffer.from(bytes);
        const value = beetSolana__namespace.publicKey.read(buffer$1, offset);
        return [value, offset + beetSolana__namespace.publicKey.byteSize];
      }
    };
  }
}

exports.BeetSerializer = BeetSerializer;
//# sourceMappingURL=BeetSerializer.cjs.map
